/*
 An example analogue meter using a ST7735 TFT LCD screen

 This example uses the hardware SPI only
 Needs Font 2 (also Font 4 if using large scale label)

 Updated by Bodmer for variable meter size
*/

// Define meter size
#define M_SIZE 1

#include <TFT_eSPI.h> // Hardware-specific library
#include <SPI.h>
#include "esp_adc_cal.h"
#include <ADS1115-Driver.h>

const int channels[4] = {
    ADS1115_MUX_AIN0_GND,
    ADS1115_MUX_AIN1_GND,
    ADS1115_MUX_AIN2_GND,
    ADS1115_MUX_AIN3_GND,
};

const int batteryPin = 34;

int battery;
int volt;
int vref = 1100;
float mag = 0.5;

#define NUM_READS 30
#define NUM_CALIBRATIONS 20

TFT_eSPI tft = TFT_eSPI(); // Invoke custom library
ADS1115 ads1115 = ADS1115(ADS1115_I2C_ADDR_GND);

#define TFT_GREY 0x5AEB
#define TFT_ORANGE 0xFD20 /* 255, 165,   0 */

uint32_t updateTime = 0; // time for next update

float tau = 0.3; // valore di esempio
// Inizializza la variabile di stato del filtro
float states[4] = {3000, 3000, 3000, 3000};
float offsets[4];
float alpha = tau / (tau + 1); // coefficiente di smoothing

// Definisci la funzione di aggiornamento del filtro
float lowpass_filter(int s, float x)
{
  states[s] = alpha * x + (1 - alpha) * states[s]; // aggiorna la variabile di stato
  return states[s];
}

void setup(void)
{
  Serial.begin(115200); // For debug
  Serial.println("SyncCarb by MP");
  tft.init();
  tft.setRotation(2);

  tft.fillScreen(TFT_BLACK);

  ads1115.reset();
  ads1115.setDeviceMode(ADS1115_MODE_SINGLE);
  ads1115.setDataRate(ADS1115_DR_860_SPS);
  ads1115.setPga(ADS1115_PGA_4_096);

  splash();
  calibrate();
  analogMeter(); // Draw analogue meter
  tft.setTextDatum(TR_DATUM);
  tft.setTextColor(TFT_WHITE); // Text colour
  updateTime = millis();       // Next update time
}

void splash()
{
  tft.setTextColor(TFT_GREEN); // Text colour
  tft.drawCentreString("SyncCarb", 68, 30, 4);
  tft.drawCentreString("by MP", 68, 60, 4);
}

void calibrate()
{
  tft.setTextColor(TFT_WHITE); // Text colour
  tft.drawCentreString("Calibrating", 68, 120, 4);
  tft.setViewport(2, 150, 130, 20);
  tft.frameViewport(TFT_WHITE, -2);
  // tft.drawRect(0, 150, 135, 30, TFT_WHITE);
  for (int i = 0; i < NUM_CALIBRATIONS; i++)
  {
    tft.fillRect(0, 0, 130 * i / NUM_CALIBRATIONS, 20, TFT_RED);
    for (int s = 0; s < 4; s++)
    {
      measureVoltage(s);
    }
  }
  for (int s = 0; s < 4; s++)
  {
    offsets[s] = states[s];
  }
  tft.resetViewport();
}

void loop()
{
    for (int i = 0; i < 4; i++)
    {
      measureVoltage(i);
      writeValue(i);
      drawGraph(i);
    }
}

void analogMeter()
{
  char buf[1];
  tft.fillScreen(TFT_BLACK);
  tft.setTextDatum(TL_DATUM);
  tft.setTextColor(TFT_GREEN); // Text colour
  // Meter outline
  for (int i = 0; i < 4; i++)
  {
    if (i == 0)
    {
      sprintf(buf, "R");
    }
    else
    {
      sprintf(buf, "%d", i + 1);
    }
    tft.drawRect(0, i * 60, 135, 30, TFT_WHITE);
    tft.fillRect(0, i * 60 + 32, 20, 25, TFT_RED);
    tft.drawString(buf, 3, i * 60 + 32, 4);
  }
}

void drawGraph(int i)
{
  int x, y;
  int deviation = states[0] - offsets[0];
  int bar = (states[i] - offsets[i] - deviation) * mag;
  int color = abs(bar) > 50 * mag ? TFT_RED : TFT_GREEN;

  tft.fillRect(1, i * 60 + 1, 133, 28, TFT_BLACK);
  if (bar > 66)
    bar = 66;
  if (bar < -65)
    bar = -65;

  if (bar == 0)
  {
    tft.fillRect(66, i * 60 + 1, bar + 1, 28, color);
  }
  else if (bar > 0)
  {
    tft.fillRect(66, i * 60 + 1, bar, 28, color);
  }
  else
  {
    tft.fillRect(66 + bar, i * 60 + 1, -bar, 28, color);
  }
}

void writeValue(int i)
{
  int x, y;
  int deviation = states[0] - offsets[0];
  int milliVolt = states[i] - offsets[i] - deviation;
  char buf[14];
  if (i == 0)
  {
    sprintf(buf, "%.2fkPa", (float)states[0] * 0.025);
  }
  else
  {
    sprintf(buf, "%.2fkPa", (float)milliVolt * 0.025);
  }
  tft.fillRect(20, i * 60 + 32, 115, 25, TFT_BLACK);
  tft.drawString(buf, 135, i * 60 + 32, 4);
}

uint16_t readValue()
{
  ads1115.startSingleConvertion();
  // delayMicroseconds(25); // The ADS1115 needs to wake up from sleep mode and usually it takes 25 uS to do that
  while (ads1115.getOperationalStatus() == 0)
    ;
  return ads1115.readConvertedValue();
}

int measureVoltage(int sensor)
{
  unsigned int raw = 0;
  unsigned int low = 5000;
  int voltage = 0;
  ads1115.setMultiplexer(channels[sensor]);
  // unsigned int ts = millis();
  for (int j = 0; j < NUM_READS; j++)
  {
    raw = readValue();
    low = min(low, raw);
  }
  // Serial.println(millis() - ts);
  low = lowpass_filter(sensor, low);
  voltage = int((float)low / 4095.0 * 3.3 * vref);
  return voltage;
}